#define __KERNEL__
#define _HURD_
#define ENONET ENETUNREACH

#define CONFIG_INET
#define CONFIG_NET

/* #undef CONFIG_INET_RARP */
/* #undef CONFIG_IP_MULTICAST */
/* #undef CONFIG_IP_FORWARD */
/* #undef CONFIG_IP_FIREWALL */
/* #undef CONFIG_IP_FIREWALL_DEBUG */
/* #undef CONFIG_IP_FIREWALL_VERBOSE */
/* #undef DEBUG_CONFIG_IP_FIREWALL */
/* #undef CONFIG_IP_ACCT */

/* #undef CONFIG_SKB_CHECK */

/* #undef CONFIG_TCP_NAGLE_OFF */
/* #undef CONFIG_AX25 */
/* #undef CONFIG_IPX */
/* #undef CONFIG_ATALK */
/* #undef CONFIG_SLAVE_BALANCING */

/* #undef CONFIG_INET_PCTCP */
/* #undef CONFIG_INET_SNARL */

/* #undef CONFIG_I_AM_A_BROKEN_BSD_WEENIE */
